#include "stdafx.h"
#include "statuscache.h"

#include "msscc.h"
#include "sccsvn.h"
#include "registry.h"

#include <time.h>
#include <sys/stat.h>
#include <stdio.h>
#include <afx.h>

#include "apr_strings.h"
#include "svn_pools.h"
#include "svn_client.h"

#include "sccutils.h"
/**
 * Constructor
 */
CStatusCache::CStatusCache(void):
m_StatusCacheHash(NULL),
m_Pool(NULL)
{
	this->initialize();
}

/**
 * Destructor
 */
CStatusCache::~CStatusCache(void)
{
	cacheData_t * data;
	apr_hash_index_t *hi;
	char *key;
	apr_ssize_t keylen;
	CString path;
	// write cache file
	if (m_StatusCacheHash)
	{
		if (apr_hash_count(m_StatusCacheHash) > 0)
		{	
			// get dir
			hi = apr_hash_first(m_Pool, m_StatusCacheHash);
			apr_hash_this(hi, (const void**)&key, &keylen, (void**)&data);

			path = GetDirectoryPart(key);
			this->writeCache(path);
		}
	}

	// destroy apr_pool
	svn_pool_destroy (m_Pool);
}

/**
 * Set a file in cache
 */
BOOL CStatusCache::Set(CString fileName, svn_wc_status_kind status, LONG sccstatus)
{
	struct _stat buffer;
	time_t fileTime=0;
	cacheData_t * data;

	if (!this->isInitialized())
	{
		this->initialize();
		if (!this->isInitialized()) return FALSE;
	}

	// get stat on file
	if (_stat(fileName, &buffer) != 0) // bad stat
	{
		// file missing ? use current time
		if (status==svn_wc_status_missing) time( &fileTime);
	} else {
		fileTime = buffer.st_mtime;
	}

	// add fileName to hash
	data = (cacheData_t * ) apr_palloc (m_Pool, sizeof (cacheData_t));
	data->status = status;
	data->fileName =  apr_pstrdup(m_Pool,fileName);
	data->fileTime = fileTime;
	data->sccstatus = sccstatus;

	apr_hash_set(m_StatusCacheHash, svn_path_local_style(fileName, m_Pool) , APR_HASH_KEY_STRING, (void*)data);

	lastAccessedData = data;
	return TRUE;
}

/**
 * Get a element from cache
 */
CStatusCache::cacheData_t * CStatusCache::Get(CString fileName)
{
	cacheData_t * data=NULL;
	if (!this->isInitialized()) return data;

	// refresh 
	if (this->isObsolete(fileName)) this->getSVNStatusForFile(fileName);

	if (lastAccessedData) 
	{
		if (fileName.Compare(lastAccessedData->fileName) == 0) return lastAccessedData;
	}

	data = (cacheData_t *) apr_hash_get(m_StatusCacheHash, fileName,APR_HASH_KEY_STRING);
	lastAccessedData = data;
	return data;
}

/**
 * true if cache is initialized
 */
BOOL CStatusCache::isInitialized() 
{
	return ((m_Pool) && (m_StatusCacheHash));
}

/**
 * Initialized data structures
 */
void CStatusCache::initialize(void)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	CsccsvnApp *theApp = (CsccsvnApp *) AfxGetApp( ); 

	// create apr_pool
	if (!m_Pool)
	{
		m_Pool = svn_pool_create(theApp->getAprPool());
		if (!m_Pool) return;
	}
	
	// create statusCacheHash
	m_StatusCacheHash = apr_hash_make(m_Pool);
	lastFullUpdate = -1;

	lastAccessedData = NULL;
}

/** 
 * Delete a file from cache
 */
BOOL CStatusCache::remove(CString fileName)
{
	if (!this->isInitialized()) return TRUE;

	// refresh 
	if (this->isObsolete(fileName)) this->getSVNStatusForFolder(fileName);

	apr_hash_set(m_StatusCacheHash, apr_pstrdup(m_Pool,fileName) , APR_HASH_KEY_STRING, NULL);
	lastAccessedData=NULL;
	return TRUE;
}

svn_wc_status_kind CStatusCache::getStatus(CString fileName)
{
	cacheData_t * data;
	data = this->Get(fileName);
	return data->status;
}

LONG CStatusCache::getSCCStatus(CString fileName)
{
	cacheData_t * data;
	data = this->Get(fileName);
	return data->sccstatus;
}

/**
 * true if empty hash table
 */
BOOL CStatusCache::isEmpty()
{
	if (!this->isInitialized()) return FALSE;

	return (apr_hash_count(m_StatusCacheHash)>0?FALSE:TRUE);
}

void CStatusCache::clean()
{
	m_StatusCacheHash = apr_hash_make(m_Pool);
}

BOOL CStatusCache::isObsolete()
{
	if (this->isEmpty()) return TRUE;
	if (this->lastFullUpdate < 0) return TRUE;	

	// cache time out
	CRegStdWORD cachetimeout(_T("Software\\SCCSVN\\UpdateCacheInSeconds"),-1, false, HKEY_LOCAL_MACHINE);
	DWORD timeout = (cachetimeout < -1?CACHETIMEOUT:cachetimeout*1000);
	return ((GetTickCount() - timeout) > this->lastFullUpdate);
}

void CStatusCache::fullUpdate()
{
	this->lastFullUpdate = GetTickCount();
}

BOOL CStatusCache::SCCPopulate(CString fileName, POPLISTFUNC pfnPopulate, LPVOID pvCallerData)
{
#ifdef DEBUG
	CString msg;
	DWORD time=GetTickCount();
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	CsccsvnApp *theApp = (CsccsvnApp *) AfxGetApp( ); 
	CContext *context = theApp->getContext();
#endif

	apr_hash_index_t *hi;
	char *key;
	apr_ssize_t keylen;
	CString path;
	LONG nStatus;
	BOOL bContinue = true;
	cacheData_t * data;

	if (!this->isInitialized()) return FALSE;

	// refresh 
	if (this->isObsolete()) this->getSVNStatusForFolder(fileName);

#ifdef DEBUG
	msg.Format("SCCPopulate,a : %d", GetTickCount() - time);
	context->print(msg, SCC_MSG_INFO);
	time = GetTickCount();
#endif
	for (hi = apr_hash_first(m_Pool, m_StatusCacheHash); hi; hi = apr_hash_next(hi)) {
		apr_hash_this(hi, (const void**)&key, &keylen, (void**)&data);

		if (data) {
			path = (char*)key;
			path.Replace("/", "\\");

			nStatus = data->sccstatus;
			if (nStatus & SCC_STATUS_CONTROLLED)
			{
				bContinue = (*pfnPopulate)(pvCallerData, true, nStatus, path);
				if (!bContinue) break;
			}
		}
    }
#ifdef DEBUG
	msg.Format("SCCPopulate,b : %d", GetTickCount() - time);
	context->print(msg, SCC_MSG_INFO);
	time = GetTickCount();
#endif 
	return bContinue;
}


BOOL CStatusCache::isObsolete(CString fileName)
{
	struct _stat buffer;
	time_t fileTime=0;
	cacheData_t * data;

	data = (cacheData_t *) apr_hash_get(m_StatusCacheHash, fileName,APR_HASH_KEY_STRING);
	if (!data) return TRUE;

	// get stat on file
	if (_stat(fileName, &buffer) == 0) fileTime = buffer.st_mtime;

	// if fileTime > data => need to update
	if ((fileTime > data->fileTime) && (data->sccstatus == SCC_STATUS_CONTROLLED))  return TRUE;

	return FALSE;
}

void CStatusCache::update(CString fileName)
{
	this->getSVNStatusForFile(fileName);
}

void CStatusCache::update(LONG nFiles, LPCSTR* lpFileNames)
{
	for (int i=0; i< nFiles; i++)
	{
		this->update(lpFileNames[i]);
	}
}

void CStatusCache::remove(LONG nFiles, LPCSTR* lpFileNames)
{
	for (int i=0; i< nFiles; i++)
	{
		this->remove(lpFileNames[i]);
	}
}

void getstatushash(void * baton, const char * path, svn_wc_status_t * status)
{
	hashbaton_t * hash = (hashbaton_t *)baton;
	svn_wc_status_t * statuscopy = svn_wc_dup_status (status, hash->pool);
	apr_hash_set (hash->hash, apr_pstrdup(hash->pool, path), APR_HASH_KEY_STRING, statuscopy);
}

BOOL CStatusCache::getSVNStatusForFolder(CString _fileName)
{
#ifdef DEBUG
	CString msg;
	DWORD time=GetTickCount();
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	CsccsvnApp *theApp = (CsccsvnApp *) AfxGetApp( ); 
	CContext *context = theApp->getContext();
#endif
	apr_hash_t *				statushash;
	svn_error_t *				m_err;
	svn_wc_status_t *			status;	
	apr_pool_t * pool; 
	CString	internalpath;
	CString resToken;
	int curPos= 0;
	svn_client_ctx_t 			* ctx;
	apr_hash_index_t *hi;
	char *key;
	apr_ssize_t keylen;
	CString path;
	LONG nStatus;
	BOOL bContinue = true;
	bool update = false; // do not change => will create connection to server, failed because no authentification defined
	bool noignore = false;
	bool bDescend = false; // recursive ?
	CString dirname;

	// check for cache file
	dirname = GetDirectoryPart(_fileName);
	if (this->readCache(dirname)) return true;

	if (!this->isInitialized()) this->initialize();
	pool = svn_pool_create (this->m_Pool);

	m_err = svn_client_create_context (&ctx, pool);
	if (m_err)
	{
		svn_pool_destroy (pool);					// free the allocated memory
		return false;
	} // if (m_err) 

	//we need to convert the path to subversion internal format
	//the internal format uses '/' instead of the windows '\'
    internalpath = svn_path_internal_style(dirname, pool);
	statushash = apr_hash_make(pool);
	svn_revnum_t youngest = SVN_INVALID_REVNUM;
	svn_opt_revision_t rev;
	rev.kind = svn_opt_revision_unspecified;
	struct hashbaton_t hashbaton;
	hashbaton.hash = statushash;
	hashbaton.pool = pool;

	m_err = svn_client_status (&youngest,
							internalpath,
							&rev,
							getstatushash,
							&hashbaton,
							bDescend,		//descend
							TRUE,		//getall
							update,		//update
							noignore,		//noignore
							ctx,
							pool);

	// Error present if function is not under version control
	if ((m_err != NULL) || (apr_hash_count(statushash) == 0))
	{
		return 0;	
	}
#ifdef DEBUG
msg.Format("getSVNStatusForFolder, status get for %d files", apr_hash_count(statushash) );
context->print(msg, SCC_MSG_INFO);
#endif
	this->clean();

	for (hi = apr_hash_first(pool, statushash); hi; hi = apr_hash_next(hi)) {
		apr_hash_this(hi, (const void**)&key, &keylen, (void**)&status);

		if (status) {
			path = (char*)key;
			path.Replace("/", "\\");

			nStatus = convertSVNStatusToSCCStatus(path, status);
			if (nStatus & SCC_STATUS_CONTROLLED)
			{
				this->Set(path, status->text_status, nStatus);
			}
		}
    }
	this->fullUpdate();
	svn_pool_destroy (pool); // free the allocated memory

	this->writeCache(dirname);
#ifdef DEBUG
msg.Format("getSVNStatusForFolder, %d files : %d", apr_hash_count(this->m_StatusCacheHash),  GetTickCount() - time);
context->print(msg, SCC_MSG_INFO);
time = GetTickCount();
#endif
	return bContinue;
}

BOOL CStatusCache::getSVNStatusForFile(CString _fileName)
{
	LPCSTR lpFileNames[1];
	lpFileNames[0] = _fileName;

	return this->getSVNStatusForFiles(1, lpFileNames);
}

BOOL CStatusCache::readCache(CString dirname)
{
//#ifdef DEBUG
	CString msg;
	DWORD time=GetTickCount();
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	CsccsvnApp *theApp = (CsccsvnApp *) AfxGetApp( ); 
	CContext *context = theApp->getContext();
//#endif
	CString cachefilename;
	CFile file;
	UINT dwRead;
	size_t len;
	char * loopfile;
	cacheData_t * data=NULL;
	CFileException e;
	char linejump;
	unsigned int nbelements;

	if (!this->isInitialized()) this->initialize();

	// get cachefilename
	cachefilename = dirname + this->getCacheName();
	
	// open file for read
	if (!file.Open(cachefilename, CFile::modeRead | CFile::typeBinary | CFile::modeCreate | CFile::modeNoTruncate))
		return false; // unable to open file

	if (file.GetLength() == 0)
	{
		file.Close();
		return false;
	}

	file.Read(&nbelements, sizeof(unsigned int));

	// write all element on hash
	do
		{
		// filename
		dwRead = file.Read(&len, sizeof(size_t));
		if (dwRead<=0) break;
		
		data = (cacheData_t *) apr_palloc(m_Pool, sizeof(cacheData_t));

		loopfile = (char *) apr_pcalloc(m_Pool, (len+1)*sizeof(char));
		dwRead = file.Read(loopfile, (int) len * sizeof(char));
		if (dwRead<=0) break;
		data->fileName = loopfile;

		// scc status
		dwRead = file.Read(&(data->sccstatus), sizeof(LONG));
		if (dwRead<=0) break;
		// time
		dwRead = file.Read(&(data->fileTime), sizeof(time_t));
		if (dwRead<=0) break;
		// svn status
		dwRead = file.Read(&(data->status), sizeof(svn_wc_status_kind));
		if (dwRead<=0) break;		
		dwRead = file.Read(&linejump, 1);

		apr_hash_set(m_StatusCacheHash, svn_path_local_style(loopfile, m_Pool) , APR_HASH_KEY_STRING, (void*)data);
	} while (dwRead>0);
	
	// close file
	file.Close();

//#ifdef DEBUG
msg.Format("readCache, %d files (%d wanted) in %d", apr_hash_count(this->m_StatusCacheHash),  nbelements, GetTickCount() - time);
context->print(msg, SCC_MSG_INFO);
time = GetTickCount();
//#endif
	
	return (apr_hash_count(m_StatusCacheHash)==nbelements);
}

BOOL CStatusCache::writeCache(CString dirname)
{
//#ifdef DEBUG
	CString msg;
	DWORD time=GetTickCount();
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	CsccsvnApp *theApp = (CsccsvnApp *) AfxGetApp( ); 
	CContext *context = theApp->getContext();
//#endif

	CString cachefilename;
	CFile file;
	cacheData_t * data;
	apr_hash_index_t *hi;
	char *key;
	apr_ssize_t keylen;
	size_t len;
	unsigned int nbelements;

	// get cachefilename
	cachefilename = dirname + this->getCacheName();

	// open file for write
	if (!file.Open(cachefilename, CFile::modeWrite | CFile::typeBinary | CFile::modeCreate))
		return false; // enable to openfile
	
	// write number of files
	nbelements = apr_hash_count(this->m_StatusCacheHash);
    file.Write(&nbelements, sizeof(unsigned int));

	// write files
	nbelements=0;
	for (hi = apr_hash_first(m_Pool, m_StatusCacheHash); hi; hi = apr_hash_next(hi)) {
		apr_hash_this(hi, (const void**)&key, &keylen, (void**)&data);
		if (keylen <= 0) continue;

		if (data) {
			// filename
			len = strlen(data->fileName);
			file.Write(&len, sizeof(size_t));
			file.Write(data->fileName, (int) strlen(data->fileName)*sizeof(char));
			// scc status
			file.Write(&(data->sccstatus), sizeof(LONG));
			// time
			file.Write(&(data->fileTime), sizeof(time_t));
			// svn status
			file.Write(&(data->status), sizeof(svn_wc_status_kind));
			// line jump
			file.Write("\n", 1);
		}
		nbelements++;
    } // for

	// close file
	file.Close();
//#ifdef DEBUG
msg.Format("writeCache, %d files (%d wanted) : %d", nbelements, apr_hash_count(this->m_StatusCacheHash),  GetTickCount() - time);
context->print(msg, SCC_MSG_INFO);
time = GetTickCount();
//#endif
	return true;
}

CString CStatusCache::getCacheName()
{
    return "\\.svn\\sccstatuscache";
}

BOOL CStatusCache::getSVNStatusForFiles(LONG nFiles, LPCSTR* lpFileNames)
{
#ifdef DEBUG
	CString msg;
	DWORD time=GetTickCount();
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	CsccsvnApp *theApp = (CsccsvnApp *) AfxGetApp( ); 
	CContext *context = theApp->getContext();
#endif

	BOOL bContinue = true;
	apr_pool_t *pool;
	svn_error_t *err;
	svn_wc_adm_access_t *adm_access;
	char *dirname; 
	const char *filename;
	svn_wc_status_t *status;
	CString path;
	LONG nStatus;
	CString _fileName;
	// check file count
	if (nFiles <= 0) return false;

	// get memory pool
	pool = svn_pool_create(m_Pool);
	if (!pool) return false;

	// get dirname from first filename, hopping there are all in same folder
	_fileName = lpFileNames[0];
	filename = svn_path_internal_style (_fileName, pool);
	if (IsDirectory(_fileName))
		dirname = (char *)filename;
	else
		dirname	= svn_path_dirname (filename, pool);

	// open workcopy
	err = svn_wc_adm_open2 (&adm_access, //svn_wc_adm_access_t **adm_access,
                               NULL, //svn_wc_adm_access_t *associated,
                               dirname, //const char *path,
                               false, //svn_boolean_t write_lock,
                               0, //int depth,
                               pool // apr_pool_t *pool
							   );
	if (err) return false;

	for (int i=0; i < nFiles; i++)
	{
		filename = svn_path_internal_style (lpFileNames[i], pool);

		err = svn_wc_status (&status, // svn_wc_status_t **status, 
							filename, //const char *path, 
							adm_access, //svn_wc_adm_access_t *adm_access,
							pool //apr_pool_t *pool
							);
		if (!err)
		{
			nStatus = convertSVNStatusToSCCStatus(path, status);
			if (nStatus & SCC_STATUS_CONTROLLED)
			{
				this->Set(filename , status->text_status, nStatus);
			}
		}
	} // for
	// close workcopy
	svn_wc_adm_close(adm_access);

	svn_pool_destroy(pool);

	return bContinue;
}
